# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import exceptions
from keystoneauth1.tests.unit import utils


class ExceptionTests(utils.TestCase):
    def test_clientexception_with_message(self):
        test_message = 'Unittest exception message.'
        exc = exceptions.ClientException(message=test_message)
        self.assertEqual(test_message, exc.message)

    def test_clientexception_with_no_message(self):
        exc = exceptions.ClientException()
        self.assertEqual(exceptions.ClientException.__name__, exc.message)

    def test_using_default_message(self):
        exc = exceptions.AuthorizationFailure()
        self.assertEqual(exceptions.AuthorizationFailure.message, exc.message)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import typing as ty
import urllib.parse

from keystoneauth1 import _utils as utils
from keystoneauth1 import access
from keystoneauth1 import discover
from keystoneauth1 import exceptions
from keystoneauth1.identity import base
from keystoneauth1.identity import v2
from keystoneauth1.identity.v3 import base as v3
from keystoneauth1 import session as ks_session


LOG = utils.get_logger(__name__)


class BaseGenericPlugin(base.BaseIdentityPlugin, metaclass=abc.ABCMeta):
    """An identity plugin that is not version dependent.

    Internally we will construct a version dependent plugin with the resolved
    URL and then proxy all calls from the base plugin to the versioned one.
    """

    auth_url: str

    def __init__(
        self,
        auth_url: ty.Optional[str] = None,
        *,
        tenant_id: ty.Optional[str] = None,
        tenant_name: ty.Optional[str] = None,
        project_id: ty.Optional[str] = None,
        project_name: ty.Optional[str] = None,
        project_domain_id: ty.Optional[str] = None,
        project_domain_name: ty.Optional[str] = None,
        domain_id: ty.Optional[str] = None,
        domain_name: ty.Optional[str] = None,
        system_scope: ty.Optional[str] = None,
        trust_id: ty.Optional[str] = None,
        default_domain_id: ty.Optional[str] = None,
        default_domain_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
    ):
        super().__init__(auth_url=auth_url, reauthenticate=reauthenticate)

        self._project_id = project_id or tenant_id
        self._project_name = project_name or tenant_name
        self._project_domain_id = project_domain_id
        self._project_domain_name = project_domain_name
        self._domain_id = domain_id
        self._domain_name = domain_name
        self._system_scope = system_scope
        self._trust_id = trust_id
        self._default_domain_id = default_domain_id
        self._default_domain_name = default_domain_name

        self._plugin: ty.Union[v2.Auth, v3.Auth, None] = None

    @abc.abstractmethod
    def create_plugin(
        self,
        session: ks_session.Session,
        version: discover._PARSED_VERSION_T,
        url: str,
        raw_status: ty.Optional[str] = None,
    ) -> ty.Union[None, v2.Auth, v3.Auth]:
        """Create a plugin from the given parameters.

        This function will be called multiple times with the version and url
        of a potential endpoint. If a plugin can be constructed that fits the
        params then it should return it. If not return None and then another
        call will be made with other available URLs.

        :param session: A session object.
        :type session: keystoneauth1.session.Session
        :param tuple version: A tuple of the API version at the URL.
        :param str url: The base URL for this version.
        :param str raw_status: The status that was in the discovery field.

        :returns: A plugin that can match the parameters or None if nothing.
        """
        return None

    @property
    def _has_domain_scope(self) -> bool:
        """Are there domain parameters.

        Domain parameters are v3 only so returns if any are set.

        :returns: True if a domain parameter is set, false otherwise.
        """
        return any(
            [
                self._domain_id,
                self._domain_name,
                self._project_domain_id,
                self._project_domain_name,
            ]
        )

    @property
    def project_domain_id(self) -> ty.Optional[str]:
        return self._project_domain_id or self._default_domain_id

    @project_domain_id.setter
    def project_domain_id(self, value: ty.Optional[str]) -> None:
        self._project_domain_id = value

    @property
    def project_domain_name(self) -> ty.Optional[str]:
        return self._project_domain_name or self._default_domain_name

    @project_domain_name.setter
    def project_domain_name(self, value: ty.Optional[str]) -> None:
        self._project_domain_name = value

    def _do_create_plugin(
        self, session: ks_session.Session
    ) -> ty.Union[v2.Auth, v3.Auth]:
        plugin = None

        try:
            disc = self.get_discovery(
                session, self.auth_url, authenticated=False
            )
        except (
            exceptions.DiscoveryFailure,
            exceptions.HttpError,
            exceptions.SSLError,
            exceptions.ConnectionError,
        ) as e:
            LOG.warning(
                'Failed to discover available identity versions when '
                'contacting %s. Attempting to parse version from URL.',
                self.auth_url,
            )

            url_parts = urllib.parse.urlparse(self.auth_url)
            path = url_parts.path.lower()

            if path.startswith('/v2.0'):
                if self._has_domain_scope:
                    raise exceptions.DiscoveryFailure(
                        'Cannot use v2 authentication with domain scope'
                    )
                plugin = self.create_plugin(session, (2, 0), self.auth_url)
            elif path.startswith('/v3'):
                plugin = self.create_plugin(session, (3, 0), self.auth_url)
            else:
                raise exceptions.DiscoveryFailure(
                    'Could not find versioned identity endpoints when '
                    'attempting to authenticate. Please check that your '
                    f'auth_url is correct. {e}'
                )

        else:
            # NOTE(jamielennox): version_data is always in oldest to newest
            # order. This is fine normally because we explicitly skip v2 below
            # if there is domain data present. With default_domain params
            # though we want a v3 plugin if available and fall back to v2 so we
            # have to process in reverse order.
            # FIXME(jamielennox): if we ever go for another version we should
            # reverse this logic as we always want to favour the newest
            # available version.
            reverse = self._default_domain_id or self._default_domain_name
            disc_data = disc.version_data(reverse=bool(reverse))

            v2_with_domain_scope = False
            for data in disc_data:
                version = data['version']

                if (
                    discover.version_match((2,), version)
                    and self._has_domain_scope
                ):
                    # NOTE(jamielennox): if there are domain parameters there
                    # is no point even trying against v2 APIs.
                    v2_with_domain_scope = True
                    continue

                plugin = self.create_plugin(
                    session,
                    version,
                    data['url'],
                    raw_status=data['raw_status'],
                )

                if plugin:
                    break
            if not plugin and v2_with_domain_scope:
                raise exceptions.DiscoveryFailure(
                    'Cannot use v2 authentication with domain scope'
                )

        if plugin:
            return plugin

        # so there were no URLs that i could use for auth of any version.
        raise exceptions.DiscoveryFailure(
            'Could not find versioned identity endpoints when attempting '
            'to authenticate. Please check that your auth_url is correct.'
        )

    def get_auth_ref(self, session: ks_session.Session) -> access.AccessInfo:
        if self._plugin:
            plugin = self._plugin
        else:
            plugin = self._do_create_plugin(session)
            self._plugin = plugin

        return plugin.get_auth_ref(session)

    @abc.abstractmethod
    def get_cache_id_elements(self) -> dict[str, ty.Optional[str]]:
        raise NotImplementedError()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import typing as ty

from keystoneauth1 import access
from keystoneauth1.identity.v3 import base
from keystoneauth1.identity.v3 import token
from keystoneauth1 import session as ks_session

__all__ = ('FederationBaseAuth',)


class _Rescoped(base.BaseAuth, metaclass=abc.ABCMeta):
    """A plugin that is always going to go through a rescope process.

    The original keystone plugins could simply pass a project or domain to
    along with the credentials and get a scoped token. For federation, K2K and
    newer mechanisms we always get an unscoped token first and then rescope.

    This is currently not public as it's generally an abstraction of a flow
    used by plugins within keystoneauth1.

    It also cannot go in base as it depends on token.Token for rescoping which
    would create a circular dependency.
    """

    rescoping_plugin = token.Token

    def get_auth_ref(self, session: ks_session.Session) -> access.AccessInfoV3:
        """Authenticate retrieve token information.

        This is a multi-step process where a client does federated authn
        receives an unscoped token.

        If an unscoped token is successfully received and scoping information
        is present then the token is rescoped to that target.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: a token data representation
        :rtype: :py:class:`keystoneauth1.access.AccessInfo`

        """
        auth_ref = self.get_unscoped_auth_ref(session)

        # narrow type
        assert auth_ref.auth_token is not None  # nosec B101

        if any(
            [
                self.trust_id,
                self.domain_id,
                self.domain_name,
                self.project_id,
                self.project_name,
                self.project_domain_id,
                self.project_domain_name,
            ]
        ):
            token_plugin = self.rescoping_plugin(
                self.auth_url,
                token=auth_ref.auth_token,
                trust_id=self.trust_id,
                domain_id=self.domain_id,
                domain_name=self.domain_name,
                project_id=self.project_id,
                project_name=self.project_name,
                project_domain_id=self.project_domain_id,
                project_domain_name=self.project_domain_name,
            )

            auth_ref = token_plugin.get_auth_ref(session)

        return auth_ref

    @abc.abstractmethod
    def get_unscoped_auth_ref(
        self, session: ks_session.Session
    ) -> access.AccessInfoV3:
        """Fetch unscoped federated token."""


class FederationBaseAuth(_Rescoped):
    """Federation authentication plugin.

    :param auth_url: URL of the Identity Service
    :type auth_url: string
    :param identity_provider: name of the Identity Provider the client
                              will authenticate against. This parameter
                              will be used to build a dynamic URL used to
                              obtain unscoped OpenStack token.
    :type identity_provider: string
    :param protocol: name of the protocol the client will authenticate
                     against.
    :type protocol: string

    """

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        *,
        trust_id: ty.Optional[str] = None,
        system_scope: ty.Optional[str] = None,
        domain_id: ty.Optional[str] = None,
        domain_name: ty.Optional[str] = None,
        project_id: ty.Optional[str] = None,
        project_name: ty.Optional[str] = None,
        project_domain_id: ty.Optional[str] = None,
        project_domain_name: ty.Optional[str] = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        super().__init__(
            auth_url=auth_url,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        self.identity_provider = identity_provider
        self.protocol = protocol

    @property
    def federated_token_url(self) -> str:
        """Full URL where authorization data is sent."""
        host = self.auth_url.rstrip('/')
        if not host.endswith('v3'):
            host += '/v3'
        values = {
            'host': host,
            'identity_provider': self.identity_provider,
            'protocol': self.protocol,
        }
        url = (
            "%(host)s/OS-FEDERATION/identity_providers/"
            "%(identity_provider)s/protocols/%(protocol)s/auth"
        )
        url = url % values

        return url



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import uuid

from oslo_utils import timeutils

from keystoneauth1 import access
from keystoneauth1 import fixture
from keystoneauth1.tests.unit import utils


class AccessV2Test(utils.TestCase):
    def test_building_unscoped_accessinfo(self):
        token = fixture.V2Token(expires='2012-10-03T16:58:01Z')

        auth_ref = access.create(body=token)

        self.assertIsInstance(auth_ref, access.AccessInfoV2)
        self.assertFalse(auth_ref.has_service_catalog())

        self.assertEqual(auth_ref.auth_token, token.token_id)
        self.assertEqual(auth_ref.username, token.user_name)
        self.assertEqual(auth_ref.user_id, token.user_id)

        self.assertEqual(auth_ref.role_ids, [])
        self.assertEqual(auth_ref.role_names, [])

        self.assertIsNone(auth_ref.tenant_name)
        self.assertIsNone(auth_ref.tenant_id)

        self.assertFalse(auth_ref.domain_scoped)
        self.assertFalse(auth_ref.project_scoped)
        self.assertFalse(auth_ref.trust_scoped)

        self.assertIsNone(auth_ref.project_domain_id)
        self.assertIsNone(auth_ref.project_domain_name)
        self.assertIsNone(auth_ref.user_domain_id)
        self.assertIsNone(auth_ref.user_domain_name)

        self.assertEqual(auth_ref.expires, token.expires)
        self.assertEqual(auth_ref.issued, token.issued)

        self.assertEqual(token.audit_id, auth_ref.audit_id)
        self.assertIsNone(auth_ref.audit_chain_id)
        self.assertIsNone(token.audit_chain_id)
        self.assertIsNone(auth_ref.bind)

    def test_will_expire_soon(self):
        token = fixture.V2Token()
        expires = timeutils.utcnow() + datetime.timedelta(minutes=5)
        token.expires = expires
        auth_ref = access.create(body=token)
        self.assertIsInstance(auth_ref, access.AccessInfoV2)
        self.assertFalse(auth_ref.will_expire_soon(stale_duration=120))
        self.assertTrue(auth_ref.will_expire_soon(stale_duration=300))
        self.assertFalse(auth_ref.will_expire_soon())

    def test_building_scoped_accessinfo(self):
        token = fixture.V2Token()
        token.set_scope()
        s = token.add_service('identity')
        s.add_endpoint('http://url')

        role_data = token.add_role()

        auth_ref = access.create(body=token)

        self.assertIsInstance(auth_ref, access.AccessInfoV2)
        self.assertTrue(auth_ref.has_service_catalog())

        self.assertEqual(auth_ref.auth_token, token.token_id)
        self.assertEqual(auth_ref.username, token.user_name)
        self.assertEqual(auth_ref.user_id, token.user_id)

        self.assertEqual(auth_ref.role_ids, [role_data['id']])
        self.assertEqual(auth_ref.role_names, [role_data['name']])

        self.assertEqual(auth_ref.tenant_name, token.tenant_name)
        self.assertEqual(auth_ref.tenant_id, token.tenant_id)

        self.assertEqual(auth_ref.tenant_name, auth_ref.project_name)
        self.assertEqual(auth_ref.tenant_id, auth_ref.project_id)

        self.assertIsNone(auth_ref.project_domain_id, 'default')
        self.assertIsNone(auth_ref.project_domain_name, 'Default')
        self.assertIsNone(auth_ref.user_domain_id, 'default')
        self.assertIsNone(auth_ref.user_domain_name, 'Default')

        self.assertTrue(auth_ref.project_scoped)
        self.assertFalse(auth_ref.domain_scoped)

        self.assertEqual(token.audit_id, auth_ref.audit_id)
        self.assertEqual(token.audit_chain_id, auth_ref.audit_chain_id)

    def test_diablo_token(self):
        diablo_token = {
            'access': {
                'token': {
                    'id': uuid.uuid4().hex,
                    'expires': '2020-01-01T00:00:10.000123Z',
                    'tenantId': 'tenant_id1',
                },
                'user': {
                    'id': 'user_id1',
                    'name': 'user_name1',
                    'roles': [{'name': 'role1'}, {'name': 'role2'}],
                },
            }
        }

        auth_ref = access.create(body=diablo_token)
        self.assertIsInstance(auth_ref, access.AccessInfoV2)

        self.assertTrue(auth_ref)
        self.assertEqual(auth_ref.username, 'user_name1')
        self.assertEqual(auth_ref.project_id, 'tenant_id1')
        self.assertEqual(auth_ref.project_name, 'tenant_id1')
        self.assertIsNone(auth_ref.project_domain_id)
        self.assertIsNone(auth_ref.project_domain_name)
        self.assertIsNone(auth_ref.user_domain_id)
        self.assertIsNone(auth_ref.user_domain_name)
        self.assertEqual(auth_ref.role_names, ['role1', 'role2'])

    def test_grizzly_token(self):
        grizzly_token = {
            'access': {
                'token': {
                    'id': uuid.uuid4().hex,
                    'expires': '2020-01-01T00:00:10.000123Z',
                },
                'user': {
                    'id': 'user_id1',
                    'name': 'user_name1',
                    'tenantId': 'tenant_id1',
                    'tenantName': 'tenant_name1',
                    'roles': [{'name': 'role1'}, {'name': 'role2'}],
                },
            }
        }

        auth_ref = access.create(body=grizzly_token)
        self.assertIsInstance(auth_ref, access.AccessInfoV2)

        self.assertEqual(auth_ref.project_id, 'tenant_id1')
        self.assertEqual(auth_ref.project_name, 'tenant_name1')
        self.assertIsNone(auth_ref.project_domain_id)
        self.assertIsNone(auth_ref.project_domain_name)
        self.assertIsNone(auth_ref.user_domain_id, 'default')
        self.assertIsNone(auth_ref.user_domain_name, 'Default')
        self.assertEqual(auth_ref.role_names, ['role1', 'role2'])

    def test_v2_roles(self):
        role_id = 'a'
        role_name = 'b'

        token = fixture.V2Token()
        token.set_scope()
        token.add_role(id=role_id, name=role_name)

        auth_ref = access.create(body=token)
        self.assertIsInstance(auth_ref, access.AccessInfoV2)

        self.assertEqual([role_id], auth_ref.role_ids)
        self.assertEqual(
            [role_id], auth_ref._data['access']['metadata']['roles']
        )
        self.assertEqual([role_name], auth_ref.role_names)
        self.assertEqual(
            [{'name': role_name}], auth_ref._data['access']['user']['roles']
        )

    def test_trusts(self):
        user_id = uuid.uuid4().hex
        trust_id = uuid.uuid4().hex

        token = fixture.V2Token(user_id=user_id, trust_id=trust_id)
        token.set_scope()
        token.add_role()

        auth_ref = access.create(body=token)
        self.assertIsInstance(auth_ref, access.AccessInfoV2)

        self.assertEqual(trust_id, auth_ref.trust_id)
        self.assertEqual(user_id, auth_ref.trustee_user_id)

        self.assertEqual(trust_id, token['access']['trust']['id'])

    def test_binding(self):
        token = fixture.V2Token()
        principal = uuid.uuid4().hex
        token.set_bind('kerberos', principal)

        auth_ref = access.create(body=token)
        self.assertIsInstance(auth_ref, access.AccessInfoV2)

        self.assertEqual({'kerberos': principal}, auth_ref.bind)

    def test_is_admin_project(self):
        token = fixture.V2Token()
        auth_ref = access.create(body=token)
        self.assertIsInstance(auth_ref, access.AccessInfoV2)
        self.assertIs(True, auth_ref.is_admin_project)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import collections
import logging
import os
import typing as ty
import warnings

import requests

from keystoneauth1 import _fair_semaphore
from keystoneauth1 import discover
from keystoneauth1 import session

if ty.TYPE_CHECKING:
    from keystoneauth1 import plugin


class _BaseAdapter:
    """An instance of a session with local variables.

    A session is a global object that is shared around amongst many clients. It
    therefore contains state that is relevant to everyone. There is a lot of
    state such as the service type and region_name that are only relevant to a
    particular client that is using the session. An adapter provides a wrapper
    of client local data around the global session object.

    version, min_version, max_version and default_microversion can all be
    given either as a string or a tuple.

    :param session: The session object to wrap.
    :type session: keystoneauth1.session.Session
    :param str service_type: The default service_type for URL discovery.
    :param str service_name: The default service_name for URL discovery.
    :param str interface: The default interface for URL discovery.
    :param str region_name: The default region_name for URL discovery.
    :param str endpoint_override:
        Always use this endpoint URL for requests for this client.
    :param version:
        The minimum version restricted to a given Major API.
        Mutually exclusive with min_version and max_version.
        (optional)
    :param auth: An auth plugin to use instead of the session one.
    :type auth: keystoneauth1.plugin.BaseAuthPlugin
    :param str user_agent: The User-Agent string to set.
    :param int connect_retries:
        The maximum number of retries that should be attempted for
        connection errors. Default None - use session default which
        is don't retry.
    :param logger:
        A logging object to use for requests that pass through this
        adapter.
    :type logger: logging.Logger
    :param dict allow:
        Extra filters to pass when discovering API versions.  (optional)
    :param dict additional_headers:
        Additional headers that should be attached to every request
        passing through the adapter. Headers of the same name specified
        per request will take priority.
    :param str client_name:
        The name of the client that created the adapter. This will be
        used to create the user_agent.
    :param str client_version:
        The version of the client that created the adapter. This will
        be used to create the user_agent.
    :param bool allow_version_hack:
        Allow keystoneauth to hack up catalog URLS to support older schemes.
        (optional, default True)
    :param str global_request_id:
        A global_request_id (in the form of ``req-$uuid``) that will be
        passed on all requests. Enables cross project request id tracking.
    :param min_version:
        The minimum major version of a given API, intended to be used as
        the lower bound of a range with max_version. Mutually exclusive with
        version. If min_version is given with no max_version it is as
        if max version is 'latest'. (optional)
    :param max_version:
        The maximum major version of a given API, intended to be used as
        the upper bound of a range with min_version. Mutually exclusive with
        version. (optional)
    :param default_microversion:
        The default microversion value to send with API requests. While
        microversions are a per-request feature, a user may know they
        want to default to sending a specific value.  (optional)
    :param int status_code_retries:
        The maximum number of retries that should be attempted for retriable
        HTTP status codes (optional, defaults to 0 - never retry).
    :param list retriable_status_codes:
        List of HTTP status codes that should be retried (optional,
        defaults to HTTP 503, has no effect when status_code_retries is 0).
    :param bool raise_exc:
        If True, requests returning failing HTTP responses will raise an
        exception; if False, the response is returned. This can be
        overridden on a per-request basis via the kwarg of the same name.
    :param float rate_limit:
        A client-side rate limit to impose on requests made through this
        adapter in requests per second. For instance, a rate_limit of 2
        means to allow no more than 2 requests per second, and a rate_limit
        of 0.5 means to allow no more than 1 request every two seconds.
        (optional, defaults to None, which means no rate limiting will be
        applied).
    :param int concurrency:
        How many simultaneous http requests this Adapter can be used for.
        (optional, defaults to None, which means no limit).
    :param float connect_retry_delay:
        Delay (in seconds) between two connect retries (if enabled).
        By default exponential retry starting with 0.5 seconds up to
        a maximum of 60 seconds is used.
    :param float status_code_retry_delay:
        Delay (in seconds) between two status code retries (if enabled).
        By default exponential retry starting with 0.5 seconds up to
        a maximum of 60 seconds is used.
    """

    client_name: ty.Optional[str] = None
    client_version: ty.Optional[str] = None

    def __init__(
        self,
        session: session.Session,
        service_type: ty.Optional[str] = None,
        service_name: ty.Optional[str] = None,
        interface: ty.Optional[str] = None,
        region_name: ty.Optional[str] = None,
        endpoint_override: ty.Optional[str] = None,
        version: ty.Optional[str] = None,
        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
        user_agent: ty.Optional[str] = None,
        connect_retries: ty.Optional[int] = None,
        logger: ty.Optional[logging.Logger] = None,
        allow: ty.Optional[dict[str, ty.Any]] = None,
        additional_headers: ty.Optional[
            collections.abc.MutableMapping[str, str]
        ] = None,
        client_name: ty.Optional[str] = None,
        client_version: ty.Optional[str] = None,
        allow_version_hack: ty.Optional[bool] = None,
        global_request_id: ty.Optional[str] = None,
        min_version: ty.Optional[str] = None,
        max_version: ty.Optional[str] = None,
        default_microversion: ty.Optional[str] = None,
        status_code_retries: ty.Optional[int] = None,
        retriable_status_codes: ty.Optional[list[int]] = None,
        raise_exc: ty.Optional[bool] = None,
        rate_limit: ty.Optional[float] = None,
        concurrency: ty.Optional[int] = None,
        connect_retry_delay: ty.Optional[float] = None,
        status_code_retry_delay: ty.Optional[float] = None,
    ):
        if version and (min_version or max_version):
            raise TypeError(
                "version is mutually exclusive with min_version and"
                " max_version"
            )
        # NOTE(jamielennox): when adding new parameters to adapter please also
        # add them to the adapter call in httpclient.HTTPClient.__init__ as
        # well as to load_adapter_from_argparse below if the argument is
        # intended to be something a user would reasonably expect to set on
        # a command line
        self.session = session
        self.service_type = service_type
        self.service_name = service_name
        self.interface = interface
        self.region_name = region_name
        self.endpoint_override = endpoint_override
        self.version = version
        self.user_agent = user_agent
        self.auth = auth
        self.connect_retries = connect_retries
        self.logger = logger
        self.allow = allow or {}
        self.additional_headers = additional_headers or {}
        self.allow_version_hack = allow_version_hack
        self.min_version = min_version
        self.max_version = max_version
        self.default_microversion = default_microversion
        self.status_code_retries = status_code_retries
        self.retriable_status_codes = retriable_status_codes
        self.connect_retry_delay = connect_retry_delay
        self.status_code_retry_delay = status_code_retry_delay
        self.raise_exc = raise_exc

        self.global_request_id = global_request_id

        if client_name:
            self.client_name = client_name
        if client_version:
            self.client_version = client_version

        rate_delay = 0.0
        if rate_limit:
            # 1 / rate converts from requests per second to delay
            # between requests needed to achieve that rate.
            rate_delay = 1.0 / rate_limit

        self._rate_semaphore = _fair_semaphore.FairSemaphore(
            concurrency, rate_delay
        )

    def _set_endpoint_filter_kwargs(self, kwargs: dict[str, object]) -> None:
        if self.service_type:
            kwargs.setdefault('service_type', self.service_type)
        if self.service_name:
            kwargs.setdefault('service_name', self.service_name)
        if self.interface:
            kwargs.setdefault('interface', self.interface)
        if self.region_name:
            kwargs.setdefault('region_name', self.region_name)
        if self.version:
            kwargs.setdefault('version', self.version)
        if self.min_version:
            kwargs.setdefault('min_version', self.min_version)
        if self.max_version:
            kwargs.setdefault('max_version', self.max_version)
        if self.allow_version_hack is not None:
            kwargs.setdefault('allow_version_hack', self.allow_version_hack)

    def _request(
        self, url: str, method: str, **kwargs: ty.Any
    ) -> requests.Response:
        endpoint_filter = kwargs.setdefault('endpoint_filter', {})
        self._set_endpoint_filter_kwargs(endpoint_filter)
        # NOTE(gmann): Convert or initlize the headers to
        # CaseInsensitiveDict to make sure headers are
        # case insensitive.
        if kwargs.get('headers'):
            kwargs['headers'] = requests.structures.CaseInsensitiveDict(
                kwargs['headers']
            )
        else:
            kwargs['headers'] = requests.structures.CaseInsensitiveDict()
        if self.endpoint_override:
            kwargs.setdefault('endpoint_override', self.endpoint_override)

        if self.auth:
            kwargs.setdefault('auth', self.auth)
        if self.user_agent:
            kwargs.setdefault('user_agent', self.user_agent)
        for arg in (
            'connect_retries',
            'status_code_retries',
            'connect_retry_delay',
            'status_code_retry_delay',
            'retriable_status_codes',
        ):
            if getattr(self, arg) is not None:
                kwargs.setdefault(arg, getattr(self, arg))
        if self.logger:
            kwargs.setdefault('logger', self.logger)
        if self.allow:
            kwargs.setdefault('allow', self.allow)
        if self.default_microversion is not None:
            kwargs.setdefault('microversion', self.default_microversion)

        if isinstance(self.session, (session.Session, Adapter)):
            # these things are unsupported by keystoneclient's session so be
            # careful with them until everyone has transitioned to ksa.
            # Allowing adapter allows adapter nesting that auth_token does.
            if self.client_name:
                kwargs.setdefault('client_name', self.client_name)
            if self.client_version:
                kwargs.setdefault('client_version', self.client_version)
            if self._rate_semaphore:
                kwargs.setdefault('rate_semaphore', self._rate_semaphore)

        else:
            warnings.warn(
                'Using keystoneclient sessions has been deprecated. '
                'Please update your software to use keystoneauth1.'
            )

        for k, v in self.additional_headers.items():
            kwargs.setdefault('headers', {}).setdefault(k, v)

        if self.global_request_id is not None:
            kwargs.setdefault('headers', {}).setdefault(
                "X-OpenStack-Request-ID", self.global_request_id
            )

        if self.raise_exc is not None:
            kwargs.setdefault('raise_exc', self.raise_exc)

        return self.session.request(url, method, **kwargs)

    def get_token(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> ty.Optional[str]:
        """Return a token as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin
                     on the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth.AuthorizationFailure: if a new
            token fetch fails.

        :returns: A valid token.
        :rtype: :class:`str`
        """
        return self.session.get_token(auth or self.auth)

    def get_endpoint(
        self,
        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
        **kwargs: ty.Any,
    ) -> ty.Optional[str]:
        """Get an endpoint as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin on
                     the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin: if a
            plugin is not available.

        :returns: An endpoint if available or None.
        :rtype: :class:`str`
        """
        if self.endpoint_override:
            return self.endpoint_override

        self._set_endpoint_filter_kwargs(kwargs)
        return self.session.get_endpoint(auth or self.auth, **kwargs)

    def get_endpoint_data(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> ty.Optional['discover.EndpointData']:
        """Get the endpoint data for this Adapter's endpoint.

        :param auth: The auth plugin to use for token. Overrides the plugin on
                     the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin: if a
            plugin is not available.
        :raises TypeError: If arguments are invalid

        :returns: Endpoint data if available or None.
        :rtype: keystoneauth1.discover.EndpointData
        """
        kwargs: dict[str, ty.Any] = {}
        self._set_endpoint_filter_kwargs(kwargs)
        if self.endpoint_override:
            kwargs['endpoint_override'] = self.endpoint_override

        return self.session.get_endpoint_data(auth or self.auth, **kwargs)

    def get_all_version_data(
        self, interface: str = 'public', region_name: ty.Optional[str] = None
    ) -> dict[str, dict[str, dict[str, list[discover.VersionData]]]]:
        """Get data about all versions of a service.

        :param interface:
            Type of endpoint to get version data for. Can be a single value
            or a list of values. A value of None indicates that all interfaces
            should be queried. (optional, defaults to public)
        :param string region_name:
            Region of endpoints to get version data for. A valueof None
            indicates that all regions should be queried. (optional, defaults
            to None)
        :returns: A dictionary keyed by region_name with values containing
            dictionaries keyed by interface with values being a list of
            :class:`~keystoneauth1.discover.VersionData`.
        """
        return self.session.get_all_version_data(
            interface=interface,
            region_name=region_name,
            service_type=self.service_type,
        )

    def get_api_major_version(
        self,
        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
        **kwargs: ty.Any,
    ) -> ty.Optional[tuple[ty.Union[int, float], ...]]:
        """Get the major API version as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin on
                     the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin: if a
            plugin is not available.

        :return: The major version of the API of the service discovered.
        :rtype: tuple or None
        """
        self._set_endpoint_filter_kwargs(kwargs)
        if self.endpoint_override:
            kwargs['endpoint_override'] = self.endpoint_override

        return self.session.get_api_major_version(auth or self.auth, **kwargs)

    def invalidate(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> bool:
        """Invalidate an authentication plugin.

        :param auth: The auth plugin to invalidate. Overrides the plugin on the
                     session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin
        """
        return self.session.invalidate(auth or self.auth)

    def get_user_id(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> ty.Optional[str]:
        """Return the authenticated user_id as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin
                     on the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
            if a new token fetch fails.
        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.

        :returns: Current `user_id` or None if not supported by plugin.
        :rtype: :class:`str`
        """
        return self.session.get_user_id(auth or self.auth)

    def get_project_id(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> ty.Optional[str]:
        """Return the authenticated project_id as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin
                     on the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
            if a new token fetch fails.
        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.

        :returns: Current `project_id` or None if not supported by plugin.
        :rtype: :class:`str`
        """
        return self.session.get_project_id(auth or self.auth)

    # TODO(efried): Move this to loading.adapter.Adapter
    @classmethod
    def register_argparse_arguments(
        cls,
        parser: argparse.ArgumentParser,
        service_type: ty.Optional[str] = None,
    ) -> None:
        """Attach arguments to a given argparse Parser for Adapters.

        :param parser: The argparse parser to attach options to.
        :type parser: argparse.ArgumentParser
        :param str service_type: Default service_type value. (optional)
        """
        adapter_group = parser.add_argument_group(
            'Service Options',
            'Options controlling the specialization of the API '
            'Connection from information found in the catalog',
        )

        adapter_group.add_argument(
            '--os-service-type',
            metavar='<name>',
            default=os.environ.get('OS_SERVICE_TYPE', service_type),
            help='Service type to request from the catalog',
        )

        adapter_group.add_argument(
            '--os-service-name',
            metavar='<name>',
            default=os.environ.get('OS_SERVICE_NAME', None),
            help='Service name to request from the catalog',
        )

        adapter_group.add_argument(
            '--os-interface',
            metavar='<name>',
            default=os.environ.get('OS_INTERFACE', 'public'),
            help='API Interface to use [public, internal, admin]',
        )

        adapter_group.add_argument(
            '--os-region-name',
            metavar='<name>',
            default=os.environ.get('OS_REGION_NAME', None),
            help='Region of the cloud to use',
        )

        adapter_group.add_argument(
            '--os-endpoint-override',
            metavar='<name>',
            default=os.environ.get('OS_ENDPOINT_OVERRIDE', None),
            help='Endpoint to use instead of the endpoint in the catalog',
        )

        adapter_group.add_argument(
            '--os-api-version',
            metavar='<name>',
            default=os.environ.get('OS_API_VERSION', None),
            help='Which version of the service API to use',
        )

    # TODO(efried): Move this to loading.adapter.Adapter
    @classmethod
    def register_service_argparse_arguments(
        cls, parser: argparse.ArgumentParser, service_type: str
    ) -> None:
        """Attach arguments to a given argparse Parser for Adapters.

        :param parser: The argparse parser to attach options to.
        :type parser: argparse.ArgumentParser
        :param str service_type: Name of a service to generate additional
                                 arguments for.
        """
        service_env = service_type.upper().replace('-', '_')
        adapter_group = parser.add_argument_group(
            f'{service_type.title()} Service Options',
            f'Options controlling the specialization of the '
            f'{service_type.title()} API Connection from information found in '
            f'the catalog',
        )

        adapter_group.add_argument(
            f'--os-{service_type}-service-type',
            metavar='<name>',
            default=os.environ.get(f'OS_{service_env}_SERVICE_TYPE', None),
            help=(
                f'Service type to request from the catalog for the '
                f'{service_type} service'
            ),
        )

        adapter_group.add_argument(
            f'--os-{service_type}-service-name',
            metavar='<name>',
            default=os.environ.get(f'OS_{service_env}_SERVICE_NAME', None),
            help=(
                f'Service name to request from the catalog for the '
                f'{service_type} service'
            ),
        )

        adapter_group.add_argument(
            f'--os-{service_type}-interface',
            metavar='<name>',
            default=os.environ.get(f'OS_{service_env}_INTERFACE', None),
            help=(
                f'API Interface to use for the {service_type} service '
                f'[public, internal, admin]'
            ),
        )

        adapter_group.add_argument(
            f'--os-{service_type}-api-version',
            metavar='<name>',
            default=os.environ.get(f'OS_{service_env}_API_VERSION', None),
            help=(
                f'Which version of the service API to use for '
                f'the {service_type} service'
            ),
        )

        adapter_group.add_argument(
            f'--os-{service_type}-endpoint-override',
            metavar='<name>',
            default=os.environ.get(
                f'OS_{service_env}_ENDPOINT_OVERRIDE', None
            ),
            help=(
                f'Endpoint to use for the {service_type} service '
                f'instead of the endpoint in the catalog'
            ),
        )


class Adapter(_BaseAdapter):
    def request(
        self, url: str, method: str, **kwargs: ty.Any
    ) -> requests.Response:
        return self._request(url, method, **kwargs)

    def get(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a GET request.

        This calls :py:meth:`.request()` with ``method`` set to ``GET``.
        """
        return self.request(url, 'GET', **kwargs)

    def head(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a HEAD request.

        This calls :py:meth:`.request()` with ``method`` set to ``HEAD``.
        """
        return self.request(url, 'HEAD', **kwargs)

    def post(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a POST request.

        This calls :py:meth:`.request()` with ``method`` set to ``POST``.
        """
        return self.request(url, 'POST', **kwargs)

    def put(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a PUT request.

        This calls :py:meth:`.request()` with ``method`` set to ``PUT``.
        """
        return self.request(url, 'PUT', **kwargs)

    def patch(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a PATCH request.

        This calls :py:meth:`.request()` with ``method`` set to ``PATCH``.
        """
        return self.request(url, 'PATCH', **kwargs)

    def delete(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a DELETE request.

        This calls :py:meth:`.request()` with ``method`` set to ``DELETE``.
        """
        return self.request(url, 'DELETE', **kwargs)


class LegacyJsonAdapter(_BaseAdapter):
    """Make something that looks like an old HTTPClient.

    A common case when using an adapter is that we want an interface similar to
    the HTTPClients of old which returned the body as JSON as well.

    You probably don't want this if you are starting from scratch.
    """

    def request(
        self, url: str, method: str, **kwargs: ty.Any
    ) -> tuple[requests.Response, object]:
        headers = kwargs.setdefault('headers', {})
        headers.setdefault('Accept', 'application/json')

        try:
            kwargs['json'] = kwargs.pop('body')
        except KeyError:
            pass

        resp = self._request(url, method, **kwargs)

        try:
            body = resp.json()
        except ValueError:
            body = None

        return resp, body

    def get(
        self, url: str, **kwargs: ty.Any
    ) -> tuple[requests.Response, object]:
        """Perform a GET request.

        This calls :py:meth:`.request()` with ``method`` set to ``GET``.
        """
        return self.request(url, 'GET', **kwargs)

    def head(
        self, url: str, **kwargs: ty.Any
    ) -> tuple[requests.Response, object]:
        """Perform a HEAD request.

        This calls :py:meth:`.request()` with ``method`` set to ``HEAD``.
        """
        return self.request(url, 'HEAD', **kwargs)

    def post(
        self, url: str, **kwargs: ty.Any
    ) -> tuple[requests.Response, object]:
        """Perform a POST request.

        This calls :py:meth:`.request()` with ``method`` set to ``POST``.
        """
        return self.request(url, 'POST', **kwargs)

    def put(
        self, url: str, **kwargs: ty.Any
    ) -> tuple[requests.Response, object]:
        """Perform a PUT request.

        This calls :py:meth:`.request()` with ``method`` set to ``PUT``.
        """
        return self.request(url, 'PUT', **kwargs)

    def patch(
        self, url: str, **kwargs: ty.Any
    ) -> tuple[requests.Response, object]:
        """Perform a PATCH request.

        This calls :py:meth:`.request()` with ``method`` set to ``PATCH``.
        """
        return self.request(url, 'PATCH', **kwargs)

    def delete(
        self, url: str, **kwargs: ty.Any
    ) -> tuple[requests.Response, object]:
        """Perform a DELETE request.

        This calls :py:meth:`.request()` with ``method`` set to ``DELETE``.
        """
        return self.request(url, 'DELETE', **kwargs)


# TODO(efried): Deprecate this in favor of
#               loading.adapter.register_argparse_arguments
def register_adapter_argparse_arguments(
    parser: argparse.ArgumentParser, service_type: ty.Optional[str] = None
) -> None:
    return Adapter.register_argparse_arguments(
        parser=parser, service_type=service_type
    )


# TODO(efried): Deprecate this in favor of
#               loading.adapter.register_service_argparse_arguments
def register_service_adapter_argparse_arguments(
    parser: argparse.ArgumentParser, service_type: str
) -> None:
    return Adapter.register_service_argparse_arguments(
        parser=parser, service_type=service_type
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import base64
import functools
import hashlib
import json
import threading
import typing as ty

from keystoneauth1 import _utils as utils
from keystoneauth1 import access
from keystoneauth1 import discover
from keystoneauth1 import exceptions
from keystoneauth1 import plugin
from keystoneauth1 import session as ks_session

LOG = utils.get_logger(__name__)


class BaseIdentityPlugin(plugin.BaseAuthPlugin, metaclass=abc.ABCMeta):
    # we count a token as valid (not needing refreshing) if it is valid for at
    # least this many seconds before the token expiry time
    MIN_TOKEN_LIFE_SECONDS: int = 120

    auth_url: ty.Optional[str]
    auth_ref: ty.Optional[access.AccessInfo]
    reauthenticate: bool

    def __init__(
        self, auth_url: ty.Optional[str] = None, reauthenticate: bool = True
    ):
        super().__init__()

        self.auth_url = auth_url
        self.auth_ref = None
        self.reauthenticate = reauthenticate

        self._lock = threading.Lock()

    @abc.abstractmethod
    def get_auth_ref(self, session: ks_session.Session) -> access.AccessInfo:
        """Obtain a token from an OpenStack Identity Service.

        This method is overridden by the various token version plugins.

        This function should not be called independently and is expected to be
        invoked via the do_authenticate function.

        This function will be invoked if the AcessInfo object cached by the
        plugin is not valid. Thus plugins should always fetch a new AccessInfo
        when invoked. If you are looking to just retrieve the current auth
        data then you should use get_access.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session

        :raises keystoneauth1.exceptions.response.InvalidResponse:
            The response returned wasn't appropriate.
        :raises keystoneauth1.exceptions.http.HttpError:
            An error from an invalid HTTP response.

        :returns: Token access information.
        :rtype: :class:`keystoneauth1.access.AccessInfo`
        """

    def get_token(self, session: ks_session.Session) -> ty.Optional[str]:
        """Return a valid auth token.

        If a valid token is not present then a new one will be fetched.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session

        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :return: A valid token.
        :rtype: string
        """
        return self.get_access(session).auth_token

    def _needs_reauthenticate(self) -> bool:
        """Return if the existing token needs to be re-authenticated.

        The token should be refreshed if it is about to expire.

        :returns: True if the plugin should fetch a new token. False otherwise.
        """
        if not self.auth_ref:
            # authentication was never fetched.
            return True

        if not self.reauthenticate:
            # don't re-authenticate if it has been disallowed.
            return False

        if self.auth_ref.will_expire_soon(self.MIN_TOKEN_LIFE_SECONDS):
            # if it's about to expire we should re-authenticate now.
            return True

        # otherwise it's fine and use the existing one.
        return False

    def get_access(
        self, session: ks_session.Session, **kwargs: ty.Any
    ) -> access.AccessInfo:
        """Fetch or return a current AccessInfo object.

        If a valid AccessInfo is present then it is returned otherwise a new
        one will be fetched.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session

        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :returns: Valid AccessInfo
        :rtype: :class:`keystoneauth1.access.AccessInfo`
        """
        # Hey Kids! Thread safety is important particularly in the case where
        # a service is creating an admin style plugin that will then proceed
        # to make calls from many threads. As a token expires all the threads
        # will try and fetch a new token at once, so we want to ensure that
        # only one thread tries to actually fetch from keystone at once.
        with self._lock:
            if self._needs_reauthenticate():
                self.auth_ref = self.get_auth_ref(session)

        # narrow type
        assert self.auth_ref is not None  # nosec B101

        return self.auth_ref

    def invalidate(self) -> bool:
        """Invalidate the current authentication data.

        This should result in fetching a new token on next call.

        A plugin may be invalidated if an Unauthorized HTTP response is
        returned to indicate that the token may have been revoked or is
        otherwise now invalid.

        :returns: True if there was something that the plugin did to
                  invalidate. This means that it makes sense to try again. If
                  nothing happens returns False to indicate give up.
        :rtype: bool
        """
        if self.auth_ref:
            self.auth_ref = None
            return True

        return False

    def get_endpoint_data(
        self,
        session: ks_session.Session,
        *,
        endpoint_override: ty.Optional[str] = None,
        discover_versions: bool = True,
        service_type: ty.Optional[str] = None,
        interface: ty.Optional[str] = None,
        region_name: ty.Optional[str] = None,
        service_name: ty.Optional[str] = None,
        allow: ty.Optional[dict[str, ty.Any]] = None,
        allow_version_hack: bool = True,
        skip_discovery: bool = False,
        min_version: ty.Optional[discover._RAW_VERSION_T] = None,
        max_version: ty.Optional[discover._RAW_VERSION_T] = None,
        **kwargs: ty.Any,
    ) -> ty.Optional[discover.EndpointData]:
        """Return a valid endpoint data for a service.

        If a valid token is not present then a new one will be fetched using
        the session and kwargs.

        version, min_version and max_version can all be given either as a
        string or a tuple.

        Valid interface types: `public` or `publicURL`,
                               `internal` or `internalURL`,
                               `admin` or 'adminURL`

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param str endpoint_override: URL to use instead of looking in the
                                      catalog. Catalog lookup will be skipped,
                                      but version discovery will be run.
                                      Sets allow_version_hack to False
                                      (optional)
        :param bool discover_versions: Whether to get version metadata from
                                       the version discovery document even
                                       if it's not neccessary to fulfill the
                                       major version request. (optional,
                                       defaults to True)
        :param string service_type: The type of service to lookup the endpoint
                                    for. This plugin will return None (failure)
                                    if service_type is not provided.
        :param interface: Type of endpoint. Can be a single value or a list
                          of values. If it's a list of values, they will be
                          looked for in order of preference. Can also be
                          `keystoneauth1.plugin.AUTH_INTERFACE` to indicate
                          that the auth_url should be used instead of the
                          value in the catalog. (optional, defaults to public)
        :param string region_name: The region the endpoint should exist in.
                                   (optional)
        :param string service_name: The name of the service in the catalog.
                                   (optional)
        :param dict allow: Extra filters to pass when discovering API
                           versions. (optional)
        :param bool allow_version_hack: Allow keystoneauth to hack up catalog
                                        URLS to support older schemes.
                                        (optional, default True)
        :param bool skip_discovery: Whether to skip version discovery even
                                    if a version has been given. This is useful
                                    if endpoint_override or similar has been
                                    given and grabbing additional information
                                    about the endpoint is not useful.
        :param min_version: The minimum version that is acceptable. Mutually
                            exclusive with version. If min_version is given
                            with no max_version it is as if max version is
                            'latest'. (optional)
        :param max_version: The maximum version that is acceptable. Mutually
                            exclusive with version. If min_version is given
                            with no max_version it is as if max version is
                            'latest'. (optional)
        :param kwargs: Ignored.

        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :return: Valid EndpointData or None if not available.
        :rtype: `keystoneauth1.discover.EndpointData` or None
        """
        allow = allow or {}

        min_version, max_version = discover._normalize_version_args(
            None, min_version, max_version, service_type=service_type
        )

        # NOTE(jamielennox): if you specifically ask for requests to be sent to
        # the auth url then we can ignore many of the checks. Typically if you
        # are asking for the auth endpoint it means that there is no catalog to
        # query however we still need to support asking for a specific version
        # of the auth_url for generic plugins.
        if interface is plugin.AUTH_INTERFACE:
            endpoint_data = discover.EndpointData(
                service_url=self.auth_url,
                service_type=service_type or 'identity',
            )
            project_id = None
        elif endpoint_override:
            # TODO(mordred) Make a code path that will look for a
            #               matching entry in the catalog if the catalog
            #               exists and fill in the interface, region_name, etc.
            #               For now, just use any information the use has
            #               provided.
            endpoint_data = discover.EndpointData(
                service_url=endpoint_override,
                catalog_url=endpoint_override,
                interface=interface,
                region_name=region_name,
                service_name=service_name,
            )
            # Setting an endpoint_override then calling get_endpoint_data means
            # you absolutely want the discovery info for the URL in question.
            # There are no code flows where this will happen for any other
            # reasons.
            allow_version_hack = False
            project_id = self.get_project_id(session)
        else:
            if not service_type:
                LOG.warning(
                    'Plugin cannot return an endpoint without '
                    'knowing the service type that is required. Add '
                    'service_type to endpoint filtering data.'
                )
                return None

            # It's possible for things higher in the stack, because of
            # defaults, to explicitly pass None.
            if not interface:
                interface = 'public'

            service_catalog = self.get_access(session).service_catalog
            project_id = self.get_project_id(session)
            # NOTE(mordred): service_catalog.url_data_for raises if it can't
            # find a match, so this will always be a valid object.
            endpoint_data = service_catalog.endpoint_data_for(
                service_type=service_type,
                interface=interface,
                region_name=region_name,
                service_name=service_name,
            )
            if not endpoint_data:
                return None

        if skip_discovery:
            return endpoint_data

        try:
            return endpoint_data.get_versioned_data(
                session,
                project_id=project_id,
                min_version=min_version,
                max_version=max_version,
                cache=self._discovery_cache,
                discover_versions=discover_versions,
                allow_version_hack=allow_version_hack,
                allow=allow,
            )
        except (
            exceptions.DiscoveryFailure,
            exceptions.HttpError,
            exceptions.ConnectionError,
        ):
            # If a version was requested, we didn't find it, return
            # None.
            if max_version or min_version:
                return None
            # If one wasn't, then the endpoint_data we already have
            # should be fine
            return endpoint_data

    def get_endpoint(
        self,
        session: ks_session.Session,
        service_type: ty.Optional[str] = None,
        interface: ty.Optional[str] = None,
        region_name: ty.Optional[str] = None,
        service_name: ty.Optional[str] = None,
        version: ty.Optional[discover._RAW_VERSION_T] = None,
        allow: ty.Optional[dict[str, ty.Any]] = None,
        allow_version_hack: bool = True,
        skip_discovery: bool = False,
        min_version: ty.Optional[discover._RAW_VERSION_T] = None,
        max_version: ty.Optional[discover._RAW_VERSION_T] = None,
        **kwargs: ty.Any,
    ) -> ty.Optional[str]:
        """Return a valid endpoint for a service.

        If a valid token is not present then a new one will be fetched using
        the session and kwargs.

        version, min_version and max_version can all be given either as a
        string or a tuple.

        Valid interface types: `public` or `publicURL`,
                               `internal` or `internalURL`,
                               `admin` or 'adminURL`

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param string service_type: The type of service to lookup the endpoint
                                    for. This plugin will return None (failure)
                                    if service_type is not provided.
        :param interface: Type of endpoint. Can be a single value or a list
                          of values. If it's a list of values, they will be
                          looked for in order of preference. Can also be
                          `keystoneauth1.plugin.AUTH_INTERFACE` to indicate
                          that the auth_url should be used instead of the
                          value in the catalog. (optional, defaults to public)
        :param string region_name: The region the endpoint should exist in.
                                   (optional)
        :param string service_name: The name of the service in the catalog.
                                   (optional)
        :param version: The minimum version number required for this
                        endpoint. (optional)
        :param dict allow: Extra filters to pass when discovering API
                           versions. (optional)
        :param bool allow_version_hack: Allow keystoneauth to hack up catalog
                                        URLS to support older schemes.
                                        (optional, default True)
        :param bool skip_discovery: Whether to skip version discovery even
                                    if a version has been given. This is useful
                                    if endpoint_override or similar has been
                                    given and grabbing additional information
                                    about the endpoint is not useful.
        :param min_version: The minimum version that is acceptable. Mutually
                            exclusive with version. If min_version is given
                            with no max_version it is as if max version is
                            'latest'. (optional)
        :param max_version: The maximum version that is acceptable. Mutually
                            exclusive with version. If min_version is given
                            with no max_version it is as if max version is
                            'latest'. (optional)

        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :return: A valid endpoint URL or None if not available.
        :rtype: string or None
        """
        # Explode `version` into min_version and max_version - everything below
        # here uses the latter rather than the former.
        min_version, max_version = discover._normalize_version_args(
            version, min_version, max_version, service_type=service_type
        )
        # Set discover_versions to False since we're only going to return
        # a URL. Fetching the microversion data would be needlessly
        # expensive in the common case. However, discover_versions=False
        # will still run discovery if the version requested is not the
        # version in the catalog.
        endpoint_data = self.get_endpoint_data(
            session,
            service_type=service_type,
            interface=interface,
            region_name=region_name,
            service_name=service_name,
            allow=allow,
            min_version=min_version,
            max_version=max_version,
            discover_versions=False,
            skip_discovery=skip_discovery,
            allow_version_hack=allow_version_hack,
            **kwargs,
        )
        return endpoint_data.url if endpoint_data else None

    def get_api_major_version(
        self,
        session: ks_session.Session,
        *,
        endpoint_override: ty.Optional[str] = None,
        service_type: ty.Optional[str] = None,
        interface: ty.Optional[str] = None,
        region_name: ty.Optional[str] = None,
        service_name: ty.Optional[str] = None,
        version: ty.Optional[str] = None,
        allow: ty.Optional[dict[str, ty.Any]] = None,
        allow_version_hack: bool = True,
        skip_discovery: bool = False,
        discover_versions: bool = False,
        min_version: ty.Optional[discover._RAW_VERSION_T] = None,
        max_version: ty.Optional[discover._RAW_VERSION_T] = None,
        **kwargs: ty.Any,
    ) -> ty.Optional[discover._PARSED_VERSION_T]:
        """Return the major API version for a service.

        If a valid token is not present then a new one will be fetched using
        the session and kwargs.

        version, min_version and max_version can all be given either as a
        string or a tuple.

        Valid interface types: `public` or `publicURL`,
                               `internal` or `internalURL`,
                               `admin` or 'adminURL`

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param str endpoint_override: URL to use for version discovery.
        :param string service_type: The type of service to lookup the endpoint
                                    for. This plugin will return None (failure)
                                    if service_type is not provided.
        :param interface: Type of endpoint. Can be a single value or a list
                          of values. If it's a list of values, they will be
                          looked for in order of preference. Can also be
                          `keystoneauth1.plugin.AUTH_INTERFACE` to indicate
                          that the auth_url should be used instead of the
                          value in the catalog. (optional, defaults to public)
        :param string region_name: The region the endpoint should exist in.
                                   (optional)
        :param string service_name: The name of the service in the catalog.
                                   (optional)
        :param version: The minimum version number required for this
                        endpoint. (optional)
        :param dict allow: Extra filters to pass when discovering API
                           versions. (optional)
        :param bool allow_version_hack: Allow keystoneauth to hack up catalog
                                        URLS to support older schemes.
                                        (optional, default True)
        :param bool skip_discovery: Whether to skip version discovery even
                                    if a version has been given. This is useful
                                    if endpoint_override or similar has been
                                    given and grabbing additional information
                                    about the endpoint is not useful.
        :param bool discover_versions: Whether to get version metadata from
                                       the version discovery document even
                                       if it's not neccessary to fulfill the
                                       major version request. Defaults to False
                                       because get_endpoint doesn't need
                                       metadata. (optional, defaults to False)
        :param min_version: The minimum version that is acceptable. Mutually
                            exclusive with version. If min_version is given
                            with no max_version it is as if max version is
                            'latest'. (optional)
        :param max_version: The maximum version that is acceptable. Mutually
                            exclusive with version. If min_version is given
                            with no max_version it is as if max version is
                            'latest'. (optional)

        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :return: The major version of the API of the service discovered.
        :rtype: tuple or None

        .. note:: Implementation notes follow. Users should not need to wrap
                  their head around these implementation notes.
                  `get_api_major_version` should do what is expected with the
                  least possible cost while still consistently returning a
                  value if possible.

        There are many cases when major version can be satisfied
        without actually calling the discovery endpoint (like when the version
        is in the url). If the user has a cloud with the versioned endpoint
        ``https://volume.example.com/v3`` in the catalog for the
        ``block-storage`` service and they do::

          client = adapter.Adapter(
              session,
              service_type='block-storage',
              min_version=2,
              max_version=3,
          )
          volume_version = client.get_api_major_version()

        The version actually be returned with no api calls other than getting
        the token. For that reason, :meth:`.get_api_major_version` first
        calls :meth:`.get_endpoint_data` with ``discover_versions=False``.

        If their catalog has an unversioned endpoint
        ``https://volume.example.com`` for the ``block-storage`` service
        and they do this::

          client = adapter.Adapter(session, service_type='block-storage')

        client is now set up to "use whatever is in the catalog". Since the
        url doesn't have a version, :meth:`.get_endpoint_data` with
        ``discover_versions=False`` will result in ``api_version=None``.
        (No version was requested so it didn't need to do the round trip)

        In order to find out what version the endpoint actually is, we must
        make a round trip. Therefore, if ``api_version`` is ``None`` after
        the first call, :meth:`.get_api_major_version` will make a second
        call to :meth:`.get_endpoint_data` with ``discover_versions=True``.

        """
        allow = allow or {}
        # Explode `version` into min_version and max_version - everything below
        # here uses the latter rather than the former.
        min_version, max_version = discover._normalize_version_args(
            version, min_version, max_version, service_type=service_type
        )
        # Using functools.partial here just to reduce copy-pasta of params
        get_endpoint_data = functools.partial(
            self.get_endpoint_data,
            session,
            endpoint_override=endpoint_override,
            service_type=service_type,
            interface=interface,
            region_name=region_name,
            service_name=service_name,
            allow=allow,
            min_version=min_version,
            max_version=max_version,
            skip_discovery=skip_discovery,
            allow_version_hack=allow_version_hack,
            **kwargs,
        )
        data = get_endpoint_data(discover_versions=discover_versions)
        if (not data or not data.api_version) and not discover_versions:
            # It's possible that no version was requested and the endpoint
            # in the catalog has no version in the URL. A version has been
            # requested, so now it's ok to run discovery.

            data = get_endpoint_data(discover_versions=True)
        if not data:
            return None
        return data.api_version

    def get_all_version_data(
        self,
        session: ks_session.Session,
        interface: str = 'public',
        region_name: ty.Optional[str] = None,
        service_type: ty.Optional[str] = None,
    ) -> dict[str, dict[str, dict[str, list[discover.VersionData]]]]:
        """Get version data for all services in the catalog.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param interface:
            Type of endpoint to get version data for. Can be a single value
            or a list of values. A value of None indicates that all interfaces
            should be queried. (optional, defaults to public)
        :param string region_name:
            Region of endpoints to get version data for. A valueof None
            indicates that all regions should be queried. (optional, defaults
            to None)
        :param string service_type:
            Limit the version data to a single service. (optional, defaults
            to None)
        :returns: A dictionary keyed by region_name with values containing
            dictionaries keyed by interface with values being a list of
            :class:`~keystoneauth1.discover.VersionData`.
        """
        service_types = discover._SERVICE_TYPES
        catalog = self.get_access(session).service_catalog
        version_data: dict[
            str, dict[str, dict[str, list[discover.VersionData]]]
        ] = {}
        endpoints_data = catalog.get_endpoints_data(
            interface=interface,
            region_name=region_name,
            service_type=service_type,
        )

        for endpoint_service_type, services in endpoints_data.items():
            if service_types.is_known(endpoint_service_type):
                endpoint_service_type = service_types.get_service_type(
                    endpoint_service_type
                )

            for service in services:
                if not service.region_name or not service.interface:
                    continue

                versions = service.get_all_version_string_data(
                    session=session, project_id=self.get_project_id(session)
                )

                if service.region_name not in version_data:
                    version_data[service.region_name] = {}
                regions = version_data[service.region_name]

                interface = service.interface.rstrip('URL')
                if interface not in regions:
                    regions[interface] = {}
                regions[interface][endpoint_service_type] = versions

        return version_data

    def get_user_id(self, session: ks_session.Session) -> ty.Optional[str]:
        return self.get_access(session).user_id

    def get_project_id(self, session: ks_session.Session) -> ty.Optional[str]:
        return self.get_access(session).project_id

    def get_sp_auth_url(
        self, session: ks_session.Session, sp_id: str
    ) -> ty.Optional[str]:
        service_providers = self.get_access(session).service_providers
        if not service_providers:
            return None
        try:
            return service_providers.get_auth_url(sp_id)
        except exceptions.ServiceProviderNotFound:
            return None

    def get_sp_url(
        self, session: ks_session.Session, sp_id: str
    ) -> ty.Optional[str]:
        service_providers = self.get_access(session).service_providers
        if not service_providers:
            return None
        try:
            return service_providers.get_sp_url(sp_id)
        except exceptions.ServiceProviderNotFound:
            return None

    def get_discovery(
        self,
        session: ks_session.Session,
        url: str,
        authenticated: ty.Optional[bool] = None,
    ) -> discover.Discover:
        """Return the discovery object for a URL.

        Check the session and the plugin cache to see if we have already
        performed discovery on the URL and if so return it, otherwise create
        a new discovery object, cache it and return it.

        This function is expected to be used by subclasses and should not
        be needed by users.

        :param session: A session object to discover with.
        :type session: keystoneauth1.session.Session
        :param str url: The url to lookup.
        :param bool authenticated: Include a token in the discovery call.
                                   (optional) Defaults to None (use a token
                                   if a plugin is installed).

        :raises keystoneauth1.exceptions.discovery.DiscoveryFailure:
            if for some reason the lookup fails.
        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :returns: A discovery object with the results of looking up that URL.
        """
        return discover.get_discovery(
            session=session,
            url=url,
            cache=self._discovery_cache,
            authenticated=authenticated,
        )

    def get_cache_id_elements(self) -> dict[str, ty.Optional[str]]:
        """Get the elements for this auth plugin that make it unique.

        As part of the get_cache_id requirement we need to determine what
        aspects of this plugin and its values that make up the unique elements.

        This should be overridden by plugins that wish to allow caching.

        :returns: The unique attributes and values of this plugin.
        :rtype: A flat dict with a str key and str or None value. This is
                required as we feed these values into a hash. Pairs where the
                value is None are ignored in the hashed id.
        """
        raise NotImplementedError()

    def get_cache_id(self) -> ty.Optional[str]:
        """Fetch an identifier that uniquely identifies the auth options.

        The returned identifier need not be decomposable or otherwise provide
        any way to recreate the plugin.

        This string MUST change if any of the parameters that are used to
        uniquely identity this plugin change. It should not change upon a
        reauthentication of the plugin.

        :returns: A unique string for the set of options
        :rtype: str or None if this is unsupported or unavailable.
        """
        try:
            elements = self.get_cache_id_elements()
        except NotImplementedError:
            return None

        hasher = hashlib.sha256()

        for k, v in sorted(elements.items()):
            if v is not None:
                hasher.update(k.encode('utf-8'))
                hasher.update(v.encode('utf-8'))

        return base64.b64encode(hasher.digest()).decode('utf-8')

    def get_auth_state(self) -> ty.Optional[str]:
        """Retrieve the current authentication state for the plugin.

        Retrieve any internal state that represents the authenticated plugin.

        This should not fetch any new data if it is not present.

        :returns: a string that can be stored or None if there is no auth state
                  present in the plugin. This string can be reloaded with
                  set_auth_state to set the same authentication.
        :rtype: str or None if no auth present.
        """
        if not self.auth_ref:
            return None

        data = {
            'auth_token': self.auth_ref.auth_token,
            'body': self.auth_ref._data,
        }

        return json.dumps(data)

    def set_auth_state(self, data: str) -> None:
        """Install existing authentication state for a plugin.

        Take the output of get_auth_state and install that authentication state
        into the current authentication plugin.
        """
        if data:
            auth_data = json.loads(data)
            self.auth_ref = access.create(
                body=auth_data['body'], auth_token=auth_data['auth_token']
            )
        else:
            self.auth_ref = None
